/* -*- C++ -*-
 * Copyright ©2004 Hugo Mills <hugo@carfax.org.uk>
 *
 * This software is distributed under the terms of the GNU GPL v2
 * For more information on the GPL, see the file COPYING or 
 * visit http://www.gnu.org/
 *
 * This software is distributed without warranty
 */

#include "magellan/configuration.h"

#include <map>
#include <set>
#include <sstream>

#include "magellan/configlexer.h"
#include "plugins.h"
#include "magellan/xfsphere.h"
#include "magellan/parsertools.h"
#include "magellan/options.h"

enum RS_Type { RS_NONE, RS_NAMED, RS_TYPED, RS_NAMED_TYPED };

/* Inserter object: 
 * Used by readsection to add an object to a map
 * Also adds to a list of plugins
 */
template <class C>
class Inserter_Map
{
public:
	Inserter_Map(std::map<std::string, C*>& mp, std::list<Plugin*>& ps)
		: target(&mp),
		  plugins(&ps)
	{ }
	virtual ~Inserter_Map() { }
	
	void insert(const std::string& key, C* value)
	{
		if(target->find(key) != target->end())
			std::cerr << "Section '" << key << "' exists already" << std::endl;
		else
		{
			value->name = key;
			(*target)[key] = value;
			plugins->push_back(value);

			if(opts->debug_parse() >= 4 || opts->debug_render() >= 2)
				std::cerr << "      " << key << " created at " << value << std::endl;
		}
	}

private:
	std::map<std::string, C*>* target;
	std::list<Plugin*>* plugins;
};

/* Inserter object:
 * Used by readsection to add an object to a simple pointer,
 * overwriting it if it was already initialised. e.g. we only expect
 * one Output object.
 */
template <class C>
class Inserter_Ptr
{
public:
	Inserter_Ptr(C& mp, std::list<Plugin*>& ps)
		: target(&mp),
		  plugins(&ps)
	{ }
	virtual ~Inserter_Ptr() { }

	void insert(const std::string& key, C value) const
	{
		if(*target != NULL)
			std::cerr << "Section '" << key << "' exists already" << std::endl;
		else
		{
			value->name = key;
			*target = value;
			plugins->push_back(value);

			if(opts->debug_parse() >= 4 || opts->debug_render() >= 2)
				std::cerr << "      " << key << " created at " << value << std::endl;
		}
	}

private:
	C* target;
	std::list<Plugin*>* plugins;
};

/* Inserter object:
 * Used by readsection to add an object to a set. e.g. in the case of
 * MDisplay objects, where we don't need to look them up, but do need
 * to be able to iterate over them.
 */
template <class C>
class Inserter_Set
{
public:
	Inserter_Set(std::set<C*>& mp, std::list<Plugin*>& ps)
		: target(&mp),
		  plugins(&ps)
	{ }
	virtual ~Inserter_Set() { }

	void insert(const std::string& key, C* value)
	{
		if(target->find(value) != target->end())
			std::cerr << "Section '" << key << "' exists already" << std::endl;
		else
		{
			value->name = key;
			target->insert(value);
			plugins->push_back(value);

			if(opts->debug_parse() >= 4 || opts->debug_render() >= 2)
				std::cerr << "      " << key << " created at " << value << std::endl;
		}
	}

private:
	std::set<C*>* target;
	std::list<Plugin*>* plugins;
};

/* Iterates over a list of plugins, and tests to see whether they will
 * accept a configuration section of the given type/subtype.
 */
template <class F>
typename F::handled_class* run_list(ConfigLexer* lex,
								   const std::string& type,
								   const std::string& subtype,
								   const std::list<F*>& pluginlist)
{
	try {
		for(typename std::list<F*>::const_iterator plug = pluginlist.begin();
			plug != pluginlist.end();
			plug++)
		{
			typename F::handled_class* m = (*plug)->parser(lex, type, subtype);
			if(m)
			{
				if(opts->debug_parse() >= 2)
					std::cerr << "  Accepted by: " << (*plug)->name 
							  << std::endl;
				lex->next();
				return m;
			}
		}
	} catch(...) {
		std::cerr << "Caught error: recovering to next section" << std::endl;
		lex->section_next();
		std::cerr << "Recovered." << std::endl;
		return NULL;
	}

	std::cerr << "No plugin accepted config '" << type << "', '" 
			  << subtype << "'. Skipping." << std::endl;
	return NULL;
}

/* Validates the current input token as the beginning of a section of
 * the given type
 */
bool check_type(ConfigLexer* lex, const char* sectype)
{
	if(lex->type() != TOKEN)
	{
		std::cerr << "Syntax error -- not a token: '"
				  << lex->value() << "' (parser_sections.h "
				  << __LINE__ << ")" << std::endl;
		// Syntax error -- not a token
		return false;
	}
	if(lex->value() != sectype)
		return false;
	lex->next();
	return true;
}

std::string get_name(ConfigLexer* lex)
{
	if(lex->type() != STRING)
	{
		// Error
		std::cerr << "FAILED" << std::endl;
		return "";
	}
	std::string rv = lex->value();
	lex->next();
	return rv;
}

bool get_types(ConfigLexer* lex,
			   std::string& type,
			   std::string& subtype)
{
	try {
		if(!readkeyedvalue(lex, "type", &type))
		{
			std::cerr << "Type not specified" << std::endl;
			return false;
		}

		if(!readkeyedvalue(lex, "subtype", &subtype))
			subtype = "";
	} catch(NotAToken natex) {
	}

	return true;
}

/* Read the first bit of a config section, including obtaining the
 * type information, if present and necessary.
 * Then find a plugin of the given type which is willing to handle that
 * configuration section.
 */
template <class Inst, class F>
bool readsection(ConfigLexer* lex,
				 enum RS_Type rst,
				 const char* sectype,
				 const std::list<F*>& pluginlist,
				 Inst& inserter)
{
	if(!check_type(lex, sectype))
		return false;

	if(opts->debug_parse() >= 1)
		std::cerr << "Reading section of class " << sectype
				  << " (line " << lex->lineno() << ")" << std::endl;

	// Read the name of the section, if it's a named section
	std::string name = "";
	if(rst == RS_NAMED || rst == RS_NAMED_TYPED)
	{
		name = get_name(lex);
		if(opts->debug_parse() >= 2)
			std::cerr << "  Named '" << name << "'" << std::endl;
	}

	// Check the opening brace
	if(lex->type() != CHAR || lex->value() != "{")
	{
		// Bitch
		return false;
	}
	lex->next();

	// Read the types, if it's a typed section
	std::string type, subtype;
	if(rst == RS_TYPED || rst == RS_NAMED_TYPED)
		if(!get_types(lex, type, subtype))
			return false;

	// Look for a plugin of the right type that's willing to handle
	// this configuration
	typename F::handled_class* obj = run_list(lex, type, subtype, pluginlist);
	if(obj == NULL)
		return false;
	// Insert the resulting config object into the relevant list
	inserter.insert(name, obj);

	return true;
}

Configuration::Configuration(ConfigLexer* lex)
	: _output(NULL)
{
	// Kick off by moving to the first token
	lex->next();

	// Temporary list of all plugins
	std::list<Plugin*> plist;

	_output = NULL;

	while(lex->type() != END)
	{
		if(lex->type() != TOKEN)
		{
			std::cerr << "Expecting a token: '" << lex->value()
					  << "' is unrecognised" << std::endl;
			lex->next();
			continue;
		}

		Inserter_Ptr<Output*> ipbm = Inserter_Ptr<Output*>(_output, plist);
		Inserter_Map<XfOrbit> imo = Inserter_Map<XfOrbit>(orbits, plist);
		Inserter_Map<XfMap> imm = Inserter_Map<XfMap>(maps, plist);
		Inserter_Map<Content> imcf = Inserter_Map<Content>(content, plist);
		Inserter_Map<XfView> imv = Inserter_Map<XfView>(viewports, plist);
		Inserter_Set<MDisplay> imdd = Inserter_Set<MDisplay>(displays, plist);

		if(readsection(lex, RS_TYPED, "output", output_plugins, ipbm)
		   || readsection(lex, RS_NAMED_TYPED, "orbit", orbit_plugins, imo)
		   || readsection(lex, RS_NAMED_TYPED, "projection", map_plugins, imm)
		   || readsection(lex, RS_NAMED, "content", content_plugins, imcf)
		   || readsection(lex, RS_NAMED, "viewport", viewport_plugins, imv)
		   || readsection(lex, RS_NONE, "display", display_plugins, imdd)
			)
		{
			// All is OK
		}
		else
		{
			// Syntax error -- unknown section
			std::cerr << "Unknown section / illegal keyword: '"
					  << lex->value() << "' (parser_config.cc "
					  << __LINE__ << ")" << std::endl;
			lex->next();
		}
	}

	// Go through all the configured objects, and attempt to bind them
	bool changed;
	bool remainingunbound;
	do {
		changed = false;
		remainingunbound = false;

		for(std::list<Plugin*>::iterator it = plist.begin();
			it != plist.end();
			++it)
		{
			if(!(*it)->isbound())
			{
				Plugin* bp = (*it)->bind(this);
				if(bp)
				{
					changed = true;
					*it = bp;
				}
			}
			if(!(*it)->isbound())
				remainingunbound = true;
		}
	} while(changed);

	if(remainingunbound)
	{
		std::cerr << "Some plugins remain unbound. Aborting." << std::endl;
		exit(1);
	}
}
